from typing import Optional

from cargo2rpm.metadata import Package, FeatureFlags, Metadata
from cargo2rpm.semver import Version


class InvalidFeatureError(Exception):
    """
    Exception which is raised when calling a method on a crate with a feature
    name as argument that is not a feature of the crate.

    This is usually the result of a packager error, for example, when
    a spec file has not been properly regenerated for a new crate version
    (which has removed a previously included feature) or after patching the
    upstream `Cargo.toml` file to remove unwanted features or optional
    dependencies.
    """

    def __init__(self, error: str):
        self.error = error
        super().__init__(error)


def msrv_from_edition(edition: str) -> Optional[str]:
    match edition:
        case "2015":
            return None
        case "2018":
            return "1.31"
        case "2021":
            return "1.56"
        case "2024":
            return "1.85"
        case _:
            msg = f"Invalid Edition: {edition}"
            raise ValueError(msg)


def buildrequires(package: Package, flags: FeatureFlags, with_dev_deps: bool) -> set[str]:
    """
    Resolves and returns RPM `BuildRequires` of the `package` crate, taking into
    account feature flags passed in the `flags` argument (which represents the
    presence of the `--all-features`, `--no-default-features, or `--features`
    CLI flags of cargo), and only includes dev-dependencies if `with_dev_deps`
    is passed as `True`.

    This happens in two stages - first, the list of enabled features (and
    whether they are used with "default" features) is resolved for all
    dependencies; then this information is used to generate the actual set of
    dependencies in RPM format.
    """

    enabled, optional_enabled, other_enabled, other_conditional = package.get_enabled_features_transitive(flags)

    normal = package.get_normal_dependencies(False)
    normal_optional = package.get_normal_dependencies(True)
    build = package.get_build_dependencies(False)
    build_optional = package.get_build_dependencies(True)
    dev = package.get_dev_dependencies()

    # keep track of dependencies and which features are enabled for them:
    # - determine union of enabled features for all dependencies
    deps_enabled_features: dict[str, set[str]] = dict()
    # - determine whether default features are enabled for all dependencies
    deps_default_features: dict[str, bool] = dict()
    # - determine optional dependencies that need to be enabled as workarounds
    workarounds: dict[str, tuple[set[str], bool]] = dict()

    # unconditionally enabled features of normal dependencies
    for name, dep in normal.items():
        features = deps_enabled_features.get(name) or set()
        features.update(dep.features)
        deps_enabled_features[name] = features

        defaults = deps_default_features.get(name) or False
        defaults = defaults or dep.uses_default_features
        deps_default_features[name] = defaults

    # unconditionally enabled features of enabled, optional, normal dependencies
    for name, dep in normal_optional.items():
        if name in optional_enabled:
            features = deps_enabled_features.get(name) or set()
            features.update(dep.features)
            deps_enabled_features[name] = features

            defaults = deps_default_features.get(name) or False
            defaults = defaults or dep.uses_default_features
            deps_default_features[name] = defaults

    # unconditionally enabled features of build-dependencies
    for name, dep in build.items():
        features = deps_enabled_features.get(name) or set()
        features.update(dep.features)
        deps_enabled_features[name] = features

        defaults = deps_default_features.get(name) or False
        defaults = defaults or dep.uses_default_features
        deps_default_features[name] = defaults

    # unconditionally enabled features of enabled, optional, build-dependencies
    for name, dep in build_optional.items():
        if name in optional_enabled:
            features = deps_enabled_features.get(name) or set()
            features.update(dep.features)
            deps_enabled_features[name] = features

            defaults = deps_default_features.get(name) or False
            defaults = defaults or dep.uses_default_features
            deps_default_features[name] = defaults

    # unconditionally enabled features of enabled dev-dependencies
    if with_dev_deps:
        for name, dep in dev.items():
            features = deps_enabled_features.get(name) or set()
            features.update(dep.features)
            deps_enabled_features[name] = features

            defaults = deps_default_features.get(name) or False
            defaults = defaults or dep.uses_default_features
            deps_default_features[name] = defaults

    # features unconditionally enabled by feature dependencies
    for name, other_features in other_enabled.items():
        features = deps_enabled_features.get(name) or set()
        features.update(other_features)
        deps_enabled_features[name] = features

        if "default" in features:
            deps_default_features[name] = True

    # features conditionally enabled by feature dependencies
    for name, other_features in other_conditional.items():
        features = deps_enabled_features.get(name) or set()
        features.update(other_features)
        deps_enabled_features[name] = features

        if name not in enabled:
            defaults = deps_default_features.get(name) or False
            if odep := normal_optional.get(name):
                defaults = defaults or odep.uses_default_features
                additional_features = odep.features
            if odep := build_optional.get(name):
                defaults = defaults or odep.uses_default_features
                additional_features = odep.features
            workarounds[name] = (set.union(other_features, additional_features), defaults)

    # collect dependencies taking into account which features are enabled
    brs = set()

    # minimum supported Rust version
    if msrv := package.rust_version:
        brs.add(f"rust >= {msrv}")
    elif msrv := msrv_from_edition(package.edition):
        brs.add(f"rust >= {msrv}")

    # normal dependencies
    for name, dep in normal.items():
        if dep.is_path_or_git():
            continue

        if deps_default_features[name]:
            brs.add(dep.to_rpm("default"))
        else:
            brs.add(dep.to_rpm(None))
        for feature in deps_enabled_features[name]:
            brs.add(dep.to_rpm(feature))

    # optional normal dependencies
    for name, dep in normal_optional.items():
        if dep.is_path_or_git():
            continue

        if name in optional_enabled:
            if deps_default_features[name]:
                brs.add(dep.to_rpm("default"))
            else:
                brs.add(dep.to_rpm(None))
            for feature in deps_enabled_features[name]:
                brs.add(dep.to_rpm(feature))

    # build-dependencies
    for name, dep in build.items():
        if dep.is_path_or_git():
            continue

        if deps_default_features[name]:
            brs.add(dep.to_rpm("default"))
        else:
            brs.add(dep.to_rpm(None))
        for feature in deps_enabled_features[name]:
            brs.add(dep.to_rpm(feature))

    # optional build-dependencies
    for name, dep in build_optional.items():
        if dep.is_path_or_git():
            continue

        if name in optional_enabled:
            if deps_default_features[name]:
                brs.add(dep.to_rpm("default"))
            else:
                brs.add(dep.to_rpm(None))
            for feature in deps_enabled_features[name]:
                brs.add(dep.to_rpm(feature))

    # dev-dependencies
    if with_dev_deps:
        for name, dep in dev.items():
            if dep.is_path_or_git():
                continue

            if deps_default_features[name]:
                brs.add(dep.to_rpm("default"))
            else:
                brs.add(dep.to_rpm(None))
            for feature in deps_enabled_features[name]:
                brs.add(dep.to_rpm(feature))

    # workarounds
    for name, (features, defaults) in workarounds.items():
        if odep := normal_optional.get(name):
            if odep.is_path_or_git():
                continue
            if defaults:
                brs.add(odep.to_rpm("default"))
            else:
                brs.add(odep.to_rpm(None))
            for feature in features:
                brs.add(odep.to_rpm(feature))
        if odep := build_optional.get(name):
            if odep.is_path_or_git():
                continue
            if defaults:
                brs.add(odep.to_rpm("default"))
            else:
                brs.add(odep.to_rpm(None))
            for feature in features:
                brs.add(odep.to_rpm(feature))

    return brs


def workspace_buildrequires(metadata: Metadata, flags: FeatureFlags, with_dev_deps: bool) -> set[str]:
    """
    Resolves and returns RPM `BuildRequires` for an entire cargo workspace.

    Prior to generating `BuildRequires` for every individual workspace member,
    intra-workspace dependencies are resolved (i.e. which features of which
    workspace member are enabled).

    This takes into account "required features" of binary targets (i.e. crates
    with "bin" or "cdylib" targets).
    """

    all_brs = set()

    member_flags = metadata.get_feature_flags_for_workspace_members(flags)
    for package in metadata.packages:
        all_brs.update(buildrequires(package, member_flags[package.name], with_dev_deps))

    return all_brs


def devel_subpackage_names(package: Package) -> set[str]:
    """
    Returns the set of subpackage names for a crate.

    If the crate does not provide a "lib" target, the set will be empty.
    Otherwise, the set of "features" of the crate is returned, with the
    implicitly defined "default" feature explicitly included.
    """

    # no feature subpackages are generated for binary-only crates
    if not package.is_lib():
        return set()

    names = package.get_feature_names()

    # the "default" feature is always implicitly defined
    if "default" not in names:
        names.add("default")

    return names


def _requires_crate(package: Package) -> set[str]:
    """
    Resolves install-time dependencies of the given crate. Used for
    automatically generating dependencies of crate packages with RPM generators.

    This only includes non-optional "normal" and "build-dependencies" of the
    crate (i.e. no enabled features or enabled optional dependencies), and
    a dependency on "cargo".
    """

    normal = package.get_normal_dependencies(False)
    build = package.get_build_dependencies(False)

    deps = set()

    # dependency on cargo is mandatory
    deps.add("cargo")

    # minimum supported Rust version
    if msrv := package.rust_version:
        deps.add(f"rust >= {msrv}")

    # normal dependencies
    for dep in normal.values():
        if dep.uses_default_features:
            deps.add(dep.to_rpm("default"))
        else:
            deps.add(dep.to_rpm(None))
        for depf in dep.features:
            deps.add(dep.to_rpm(depf))

    # build-dependencies
    for dep in build.values():
        if dep.uses_default_features:
            deps.add(dep.to_rpm("default"))
        else:
            deps.add(dep.to_rpm(None))
        for depf in dep.features:
            deps.add(dep.to_rpm(depf))

    return deps


def _requires_feature(package: Package, feature: str) -> set[str]:
    """
    Resolves install-time dependencies of the given crate feature.

    This includes optional "normal" and "build-dependencies" of the
    crate that are specified as dependencies of the given feature, a
    dependency on the "main" crate package, and a dependency on "cargo".

    Raises an `InvalidFeatureError` if the given feature is not a feature
    of the crate.
    """

    if feature != "default" and feature not in package.get_feature_names():
        raise InvalidFeatureError(f"Unknown feature: {feature}")

    deps = set()

    # dependency on cargo is mandatory
    deps.add("cargo")

    if feature == "default" and "default" not in package.get_feature_names():
        # default feature is implicitly defined but empty
        deps.add(package.to_rpm_dependency(None))
        return deps

    feature_deps = package.features[feature]

    normal = package.get_normal_dependencies(False)
    normal_optional = package.get_normal_dependencies(True)
    build_optional = package.get_build_dependencies(True)

    # always add a dependency on the main crate
    deps.add(package.to_rpm_dependency(None))

    for fdep in feature_deps:
        if fdep.startswith("dep:"):
            # optional dependency
            name = fdep.removeprefix("dep:")

            found = False
            if dep := normal_optional.get(name):
                # optional normal dependency
                found = True
                if dep.uses_default_features:
                    deps.add(dep.to_rpm("default"))
                else:
                    deps.add(dep.to_rpm(None))
                for depf in dep.features:
                    deps.add(dep.to_rpm(depf))

            if dep := build_optional.get(name):
                # optional build-dependency
                found = True
                if dep.uses_default_features:
                    deps.add(dep.to_rpm("default"))
                else:
                    deps.add(dep.to_rpm(None))
                for depf in dep.features:
                    deps.add(dep.to_rpm(depf))

            if not found:  # pragma nocover
                raise InvalidFeatureError(f"No optional dependency found with name {name!r}.")

        elif "/" in fdep and "?/" not in fdep:
            # dependency with specified feature
            name, feat = fdep.split("/")

            # implicitly enabled optional dependency
            if dep := normal_optional.get(name):
                deps.add(dep.to_rpm(feat))
                if dep.uses_default_features:
                    deps.add(dep.to_rpm("default"))
                else:
                    deps.add(dep.to_rpm(None))
                for depf in dep.features:
                    deps.add(dep.to_rpm(depf))

            if dep := build_optional.get(name):
                deps.add(dep.to_rpm(feat))
                if dep.uses_default_features:
                    deps.add(dep.to_rpm("default"))
                else:
                    deps.add(dep.to_rpm(None))
                for depf in dep.features:
                    deps.add(dep.to_rpm(depf))

            # normal dependency
            if dep := normal.get(name):
                deps.add(dep.to_rpm(feat))

        elif "?/" in fdep:
            # conditionally enabled dependency feature
            name, feat = fdep.split("?/")

            if dep := normal_optional.get(name):
                deps.add(f"{dep.to_rpm(feat)}")
                if dep.uses_default_features:
                    deps.add(dep.to_rpm("default"))
                else:
                    deps.add(dep.to_rpm(None))
                for depf in dep.features:
                    deps.add(dep.to_rpm(depf))

            if dep := build_optional.get(name):
                deps.add(f"{dep.to_rpm(feat)}")
                if dep.uses_default_features:
                    deps.add(dep.to_rpm("default"))
                else:
                    deps.add(dep.to_rpm(None))
                for depf in dep.features:
                    deps.add(dep.to_rpm(depf))

        else:
            # dependency on a feature of the current crate
            if fdep not in package.get_feature_names():  # pragma nocover
                raise InvalidFeatureError(f"Invalid feature dependency (not a feature name): {fdep!r}")

            deps.add(package.to_rpm_dependency(fdep))

    return deps


def requires(package: Package, feature: Optional[str]) -> set[str]:  # pragma nocover
    if feature is None:
        return _requires_crate(package)
    else:
        return _requires_feature(package, feature)


def _provides_crate(package: Package) -> str:
    """
    Returns the standardized identifier for the "main" crate subpackage
    (i.e. `crate(foo) = x.y.z`), which is used for the automatic generation of
    "virtual Provides".
    """

    rpm_version = Version.parse(package.version).to_rpm()
    return f"crate({package.name}) = {rpm_version}"


def _provides_feature(package: Package, feature: str) -> str:
    """
    Returns a standardized identifier for the "feature" crate subpackages
    (i.e. `crate(foo/bar) = x.y.z`), which is used for the automatic generation
    of "virtual Provides".
    """

    if feature != "default" and feature not in package.get_feature_names():
        raise InvalidFeatureError(f"Unknown feature: {feature}")

    rpm_version = Version.parse(package.version).to_rpm()
    return f"crate({package.name}/{feature}) = {rpm_version}"


def provides(package: Package, feature: Optional[str]) -> str:  # pragma nocover
    if feature is None:
        return _provides_crate(package)
    else:
        return _provides_feature(package, feature)



from cargo2rpm.semver import PreRelease

import pytest


@pytest.mark.parametrize(
    "string,expected",
    [
        ("alpha", PreRelease(["alpha"])),
        ("alpha.beta", PreRelease(["alpha", "beta"])),
        ("alpha.beta.1", PreRelease(["alpha", "beta", 1])),
        ("alpha.1", PreRelease(["alpha", 1])),
        ("alpha0.valid", PreRelease(["alpha0", "valid"])),
        ("alpha.0valid", PreRelease(["alpha", "0valid"])),
        ("alpha-a.b-c-somethinglong", PreRelease(["alpha-a", "b-c-somethinglong"])),
        ("rc.1", PreRelease(["rc", 1])),
        ("DEV-SNAPSHOT", PreRelease(["DEV-SNAPSHOT"])),
        ("SNAPSHOT-123", PreRelease(["SNAPSHOT-123"])),
        ("alpha.1227", PreRelease(["alpha", 1227])),
        ("---RC-SNAPSHOT.12.9.1--.12", PreRelease(["---RC-SNAPSHOT", 12, 9, "1--", 12])),
        ("---R-S.12.9.1--.12", PreRelease(["---R-S", 12, 9, "1--", 12])),
        ("---RC-SNAPSHOT.12.9.1--.12", PreRelease(["---RC-SNAPSHOT", 12, 9, "1--", 12])),
        ("0A.is.legal", PreRelease(["0A", "is", "legal"])),
    ],
    ids=repr,
)
def test_parse_prerelease(string: str, expected: PreRelease):
    assert PreRelease.parse(string) == expected



from cargo2rpm.metadata import FeatureFlags

import pytest


def test_feature_flags_invalid():
    with pytest.raises(ValueError) as exc:
        FeatureFlags(all_features=True, features=["default"])
    assert "Cannot specify both '--all-features' and '--features'." in str(exc.value)

    with pytest.raises(ValueError) as exc:
        FeatureFlags(no_default_features=True, all_features=True)
    assert "Cannot specify both '--all-features' and '--no-default-features'." in str(exc.value)



from cargo2rpm.semver import VERSION_REGEX, VERSION_REQ_REGEX
from cargo2rpm.utils import short_repr

import pytest


@pytest.mark.parametrize(
    "string,expected",
    [
        ("1.0.0-alpha", dict(major="1", minor="0", patch="0", pre="alpha", build=None)),
        ("1.0.0-alpha.1", dict(major="1", minor="0", patch="0", pre="alpha.1", build=None)),
        ("1.0.0-0.3.7", dict(major="1", minor="0", patch="0", pre="0.3.7", build=None)),
        ("1.0.0-x.7.z.92", dict(major="1", minor="0", patch="0", pre="x.7.z.92", build=None)),
        ("1.0.0-x-y-z.--", dict(major="1", minor="0", patch="0", pre="x-y-z.--", build=None)),
        ("1.0.0-alpha+001", dict(major="1", minor="0", patch="0", pre="alpha", build="001")),
        ("1.0.0+20130313144700", dict(major="1", minor="0", patch="0", pre=None, build="20130313144700")),
        ("1.0.0-beta+exp.sha.5114f85", dict(major="1", minor="0", patch="0", pre="beta", build="exp.sha.5114f85")),
        ("1.0.0+21AF26D3----117B344092BD", dict(major="1", minor="0", patch="0", pre=None, build="21AF26D3----117B344092BD")),
    ],
    ids=short_repr,
)
def test_version_regex(string: str, expected: dict):
    match = VERSION_REGEX.match(string)
    assert match is not None
    assert match.groupdict() == expected


@pytest.mark.parametrize(
    "string,expected",
    [
        ("1", dict(op=None, major="1", minor=None, patch=None, pre=None)),
        ("1.2", dict(op=None, major="1", minor="2", patch=None, pre=None)),
        ("1.2.3", dict(op=None, major="1", minor="2", patch="3", pre=None)),
        ("1.2.3-alpha.1", dict(op=None, major="1", minor="2", patch="3", pre="alpha.1")),
        ("1.2.3-alpha2", dict(op=None, major="1", minor="2", patch="3", pre="alpha2")),
        ("=0", dict(op="=", major="0", minor=None, patch=None, pre=None)),
        ("=0.11", dict(op="=", major="0", minor="11", patch=None, pre=None)),
        ("=0.1.17", dict(op="=", major="0", minor="1", patch="17", pre=None)),
        ("=0.2.37-alpha.1", dict(op="=", major="0", minor="2", patch="37", pre="alpha.1")),
        ("=0.2.37-alpha2", dict(op="=", major="0", minor="2", patch="37", pre="alpha2")),
        (">1", dict(op=">", major="1", minor=None, patch=None, pre=None)),
        (">1.2", dict(op=">", major="1", minor="2", patch=None, pre=None)),
        (">1.2.3", dict(op=">", major="1", minor="2", patch="3", pre=None)),
        (">1.2.3-alpha.1", dict(op=">", major="1", minor="2", patch="3", pre="alpha.1")),
        (">1.2.3-alpha2", dict(op=">", major="1", minor="2", patch="3", pre="alpha2")),
        (">=0", dict(op=">=", major="0", minor=None, patch=None, pre=None)),
        (">=0.11", dict(op=">=", major="0", minor="11", patch=None, pre=None)),
        (">=0.1.17", dict(op=">=", major="0", minor="1", patch="17", pre=None)),
        (">=0.2.37-alpha.1", dict(op=">=", major="0", minor="2", patch="37", pre="alpha.1")),
        (">=0.2.37-alpha2", dict(op=">=", major="0", minor="2", patch="37", pre="alpha2")),
        ("<1", dict(op="<", major="1", minor=None, patch=None, pre=None)),
        ("<1.2", dict(op="<", major="1", minor="2", patch=None, pre=None)),
        ("<1.2.3", dict(op="<", major="1", minor="2", patch="3", pre=None)),
        ("<1.2.3-alpha.1", dict(op="<", major="1", minor="2", patch="3", pre="alpha.1")),
        ("<1.2.3-alpha2", dict(op="<", major="1", minor="2", patch="3", pre="alpha2")),
        ("<=0", dict(op="<=", major="0", minor=None, patch=None, pre=None)),
        ("<=0.11", dict(op="<=", major="0", minor="11", patch=None, pre=None)),
        ("<=0.1.17", dict(op="<=", major="0", minor="1", patch="17", pre=None)),
        ("<=0.2.37-alpha.1", dict(op="<=", major="0", minor="2", patch="37", pre="alpha.1")),
        ("<=0.2.37-alpha2", dict(op="<=", major="0", minor="2", patch="37", pre="alpha2")),
        ("~1", dict(op="~", major="1", minor=None, patch=None, pre=None)),
        ("~1.2", dict(op="~", major="1", minor="2", patch=None, pre=None)),
        ("~1.2.3", dict(op="~", major="1", minor="2", patch="3", pre=None)),
        ("~1.2.3-alpha.1", dict(op="~", major="1", minor="2", patch="3", pre="alpha.1")),
        ("~1.2.3-alpha2", dict(op="~", major="1", minor="2", patch="3", pre="alpha2")),
        ("^0", dict(op="^", major="0", minor=None, patch=None, pre=None)),
        ("^0.11", dict(op="^", major="0", minor="11", patch=None, pre=None)),
        ("^0.1.17", dict(op="^", major="0", minor="1", patch="17", pre=None)),
        ("^0.2.37-alpha.1", dict(op="^", major="0", minor="2", patch="37", pre="alpha.1")),
        ("^0.2.37-alpha2", dict(op="^", major="0", minor="2", patch="37", pre="alpha2")),
        ("1.*", dict(op=None, major="1", minor="*", patch=None, pre=None)),
        ("1.2.*", dict(op=None, major="1", minor="2", patch="*", pre=None)),
        ("1.*.*", dict(op=None, major="1", minor="*", patch="*", pre=None)),
    ],
    ids=short_repr,
)
def test_version_req_regex(string: str, expected: dict):
    match = VERSION_REQ_REGEX.match(string)
    assert match is not None
    assert match.groupdict() == expected



import json
import re
import subprocess
import sys
import textwrap
from typing import Optional

from cargo2rpm import CARGO
from cargo2rpm.semver import Comparator, Op, Version, VersionReq


class FeatureFlags:
    """
    Collection of flags that affect feature and dependency resolution
    (i.e. `--all-features`, `--no-default-features`, and `--features foo,bar`).

    Raises a `ValueError` during initialization if arguments that are
    incompatible with each other are passed.

    Passing no arguments is equivalent to passing no command-line flags to
    cargo, i.e. the "default" feature is enabled.
    """

    def __init__(self, all_features: bool = False, no_default_features: bool = False, features: Optional[list[str]] = None):
        if features is None:
            features = []

        if all_features and features:
            raise ValueError("Cannot specify both '--all-features' and '--features'.")

        if all_features and no_default_features:
            raise ValueError("Cannot specify both '--all-features' and '--no-default-features'.")

        self.all_features = all_features
        self.no_default_features = no_default_features
        self.features = features

    def __repr__(self):
        parts = []

        if self.all_features:
            parts.append("all_features")
        if self.no_default_features:
            parts.append("no_default_features")
        if self.features:
            parts.append(f"features=[{', '.join(self.features)}]")

        if parts:
            string = ", ".join(parts)
            return f"[{string}]"
        else:
            return "[]"

    def __eq__(self, other):
        if not isinstance(other, FeatureFlags):
            return False  # pragma nocover

        return (
            self.all_features == other.all_features
            and self.no_default_features == other.no_default_features
            and set(self.features) == set(other.features)
        )


class Dependency:
    def __init__(self, data):
        self._data = data

    def __repr__(self):
        return repr(self._data)

    @property
    def name(self) -> str:
        return self._data["name"]

    @property
    def req(self) -> str:
        return self._data["req"]

    @property
    def kind(self) -> Optional[str]:
        return self._data["kind"]

    @property
    def rename(self) -> Optional[str]:
        return self._data["rename"]

    @property
    def optional(self) -> bool:
        return self._data["optional"]

    @property
    def uses_default_features(self) -> bool:
        return self._data["uses_default_features"]

    @property
    def features(self) -> list[str]:
        return self._data["features"]

    @property
    def target(self) -> Optional[str]:
        return self._data["target"]

    @property
    def path(self) -> Optional[str]:
        return self._data.get("path")

    @property
    def source(self) -> Optional[str]:
        return self._data.get("source")

    def is_path_or_git(self) -> bool:
        if self.path:
            return True
        if source := self.source:
            if source.startswith("git+"):
                return True
        return False

    def to_rpm(self, feature: Optional[str]) -> str:
        """
        Formats this crate dependency as an RPM dependency string.
        """

        assert self.path is None, "Attempt to generate an RPM dependency for a path dependency!"

        req = VersionReq.parse(self.req)
        return req.to_rpm(self.name, feature)


class Target:
    def __init__(self, data):
        self._data = data

    def __repr__(self):
        return repr(self._data)

    @property
    def name(self) -> str:
        return self._data["name"]

    @property
    def kind(self) -> list[str]:
        return self._data["kind"]

    @property
    def crate_types(self) -> list[str]:
        return self._data["crate_types"]

    @property
    def required_features(self) -> list[str]:
        return self._data.get("required-features", None) or list()


class Package:
    def __init__(self, data):
        self._data = data

    def __repr__(self):
        return repr(self._data)

    @property
    def name(self) -> str:
        return self._data["name"]

    @property
    def version(self) -> str:
        return self._data["version"]

    @property
    def license(self) -> Optional[str]:
        return self._data["license"]

    @property
    def license_file(self) -> Optional[str]:
        return self._data["license_file"]

    @property
    def description(self) -> Optional[str]:
        return self._data["description"]

    @property
    def dependencies(self) -> list[Dependency]:
        return [Dependency(dependency) for dependency in self._data["dependencies"]]

    @property
    def targets(self) -> list[Target]:
        return [Target(target) for target in self._data["targets"]]

    @property
    def features(self) -> dict[str, list[str]]:
        return self._data["features"]

    @property
    def manifest_path(self) -> str:
        return self._data["manifest_path"]

    @property
    def rust_version(self) -> Optional[str]:
        return self._data["rust_version"]

    @property
    def edition(self) -> str:
        return self._data["edition"]

    @property
    def homepage(self) -> Optional[str]:
        return self._data["homepage"]

    @property
    def repository(self) -> Optional[str]:
        return self._data["repository"]

    def get_feature_names(self) -> set[str]:
        return set(self.features.keys())

    def get_normal_dependencies(self, optional: bool) -> dict[str, Dependency]:
        """
        Returns a dictionary that maps "normal" dependencies (i.e. not build- or
        dev-dependencies) from their possibly renamed name (i.e. how they can be
        referenced in feature dependencies) to the dependency objects
        themselves.
        """

        normal = filter(lambda d: d.kind is None and d.optional == optional, self.dependencies)
        return {d.rename if d.rename else d.name: d for d in normal}

    def get_build_dependencies(self, optional: bool) -> dict[str, Dependency]:
        """
        Returns a dictionary that maps build-dependencies from their possibly
        renamed name (i.e. how they can be referenced in feature dependencies)
        to the dependency objects themselves.
        """

        build = filter(lambda d: d.kind == "build" and d.optional == optional, self.dependencies)
        return {d.rename if d.rename else d.name: d for d in build}

    def get_dev_dependencies(self) -> dict[str, Dependency]:
        """
        Returns a dictionary that maps dev-dependencies from their possibly
        renamed name (i.e. how they can be referenced in feature dependencies)
        to the dependency objects themselves.
        """

        dev = filter(lambda d: d.kind == "dev", self.dependencies)
        return {d.rename if d.rename else d.name: d for d in dev}

    def to_rpm_dependency(self, feature: Optional[str]) -> str:
        """
        Returns an RPM dependency string that represents a dependency on this
        crate with an exact requirement on its current version.
        """

        ver = Version.parse(self.version)
        req = VersionReq([Comparator(Op.EXACT, ver.major, ver.minor, ver.patch, ver.pre)])
        return req.to_rpm(self.name, feature)

    def get_description(self) -> Optional[str]:
        """
        Returns a reformatted version of the package description with lines
        wrapped to 72 characters.
        """

        if self.description is None:
            return None

        # reformat contents so paragraphs become lines
        paragraphs = self.description.replace("\n\n", "\r").replace("\n", " ").replace("\r", "\n").strip()

        # ensure description starts with a capital letter
        if not paragraphs[0].isupper():
            paragraphs = paragraphs[0].upper() + paragraphs[1:]

        # ensure description ends with a full stop
        if not paragraphs.endswith("."):
            paragraphs += "."

        # return contents wrapped to 72 columns
        return "\n".join(textwrap.wrap(paragraphs, 72))

    def get_summary(self) -> Optional[str]:
        """
        Returns a shortened version of the package description based on a few
        heuristics.
        """

        if not self.description:
            return None

        # replace markdown markup (i.e. code fences)
        description = self.description.replace("`", "")

        # replace common phrases like "this is a" or "this {crate} provides an"
        stripped = re.sub(
            r"^((a|an|this)\s+)?(crate\s+)?((is|provides)\s+)?((a|an|the)\s+)?",
            "",
            description,
            flags=re.IGNORECASE,
        )

        # if stripped description still contains multiple lines, merge them
        stripped = re.sub(r"(\n+)", " ", stripped).strip()

        # ensure summary starts with a capital letter
        if not stripped[0].isupper():
            stripped = stripped[0].upper() + stripped[1:]

        # if length is already short enough, reformat as one line and return
        if len(stripped) <= 72:
            return stripped.removesuffix(".")

        # use some heuristics to determine phrase boundaries
        if (brace := stripped.find(" (")) != -1:
            return stripped[0:brace].removesuffix(".")
        if (period := stripped.find(". ")) != -1:
            return stripped[0:period].removesuffix(".")
        if (semicolon := stripped.find("; ")) != -1:
            return stripped[0:semicolon].removesuffix(".")

        # none of the heuristics matched:
        # fall back to returning the stripped description even if it's too long
        return stripped.removesuffix(".")

    def is_bin(self) -> bool:
        """
        Returns `True` if any of the build targets of this crate are a "proper"
        binary target (i.e. crate type is `bin`, but not `example`, `test`, or
        `bench`).
        """

        for target in self.targets:
            if "bin" in target.kind:
                return True
        return False

    def is_lib(self) -> bool:
        """
        Returns `True` if any of the build targets of this crate are either a
        `(r)lib` target with `(r)lib` crate type, or a `proc-macro` target with
        `proc-macro` crate type.
        """

        for target in self.targets:
            if "lib" in target.kind and "lib" in target.crate_types:
                return True
            if "rlib" in target.kind and "rlib" in target.crate_types:
                return True
            if "proc-macro" in target.kind and "proc-macro" in target.crate_types:
                return True
        return False

    def is_cdylib(self) -> bool:
        """
        Returns `True` if any of the build targets of this crate are a `lib`
        target with `cdylib` crate type.
        """

        for target in self.targets:
            if "cdylib" in target.kind and "cdylib" in target.crate_types:
                return True
        return False

    def get_binaries(self) -> set[str]:
        """
        Returns the set of "proper" binary build targets of this crate (i.e.
        the crate type is `bin`, but not `example`, `test`, or `bench`).
        """

        bins = set()
        for target in self.targets:
            if "bin" in target.kind:
                bins.add(target.name)
        return bins

    def get_enabled_features_transitive(self, flags: FeatureFlags) -> tuple[set[str], set[str], dict[str, set[str]], dict[str, set[str]]]:
        """
        Resolves the transitive closure of enabled features, enabled optional
        dependencies, enabled features of (optional or non-optional)
        dependencies, and conditionally enabled features of optional
        dependencies, taking feature flags into account.
        """

        # get names of all optional dependencies
        optional_names = set(self.get_normal_dependencies(True).keys()).union(set(self.get_build_dependencies(True).keys()))

        # collect enabled features of this crate
        enabled: set[str] = set()
        # collect enabled optional dependencies
        optional_enabled: set[str] = set()
        # collect enabled features of other crates
        other_enabled: dict[str, set[str]] = dict()
        # collect conditionally enabled features of other crates
        other_conditional: dict[str, set[str]] = dict()

        # process arguments
        feature_names = self.get_feature_names()

        if not flags.no_default_features and "default" not in flags.features and "default" in feature_names:
            enabled.add("default")

        if flags.all_features:
            for feature in feature_names:
                enabled.add(feature)

        for feature in flags.features:
            enabled.add(feature)

        # calculate transitive closure of enabled features
        while True:
            new = set()

            for feature in enabled:
                deps = self.features[feature]

                for dep in deps:
                    # named optional dependency
                    if dep.startswith("dep:"):
                        name = dep.removeprefix("dep:")
                        optional_enabled.add(name)
                        continue

                    # dependency/feature
                    if "/" in dep and "?/" not in dep:
                        name, feat = dep.split("/")

                        # using "foo/bar" in feature dependencies implicitly
                        # also enables the optional dependency "foo":
                        if name in optional_names:
                            optional_enabled.add(name)

                        if name in other_enabled.keys():
                            other_enabled[name].add(feat)
                        else:
                            other_enabled[name] = {feat}
                        continue

                    # dependency?/feature
                    if "?/" in dep:
                        name, feat = dep.split("?/")
                        if name in other_conditional.keys():
                            other_conditional[name].add(feat)
                        else:
                            other_conditional[name] = {feat}
                        continue

                    # feature name
                    if dep not in enabled:
                        new.add(dep)

            # continue until set of enabled "proper" features no longer changes
            if new:
                enabled.update(new)
            else:
                break

        return enabled, optional_enabled, other_enabled, other_conditional


class Metadata:
    """
    Representation of top-level crate metadata, i.e. the entire JSON dump
    produced by calling `cargo metadata --format-version 1`. The format of this
    data is guaranteed to be stable.

    The format of the data is the same whether run against `Cargo.toml` from an
    isolated crate or against a `Cargo.toml` manifest that defines a cargo
    workspace. The only difference is that for a single crate, the list of
    packages will be of length one, and for a workspace, the list of packages
    will (in general) be two or larger.
    """

    def __init__(self, data):
        self._data = data

    def __repr__(self):
        return repr(self._data)

    @staticmethod
    def from_json(data: str) -> "Metadata":
        """
        Loads JSON dump from input data and returns a `Metadata` object.

        This method is used for loading JSON dumps for test input, as it does
        not require any other files (i.e. crate sources) to be present.
        """

        return Metadata(json.loads(data))

    @staticmethod
    def from_cargo(path: str) -> "Metadata":  # pragma nocover
        """
        Runs `cargo metadata` with the appropriate flags against the
        `Cargo.toml` manifest at `path`.

        This method only returns correct results when run against `Cargo.toml`
        files included in complete crate sources. Otherwise, automatic target
        discovery (i.e. the default behaviour of `autobins`, `autoexamples`,
        `autotests`, and `autobenches`) will result in missing build targets,
        and missing source files for explicitly specified `bin`, `example`,
        `test`, or `bench` targets will cause errors.
        """

        ret = subprocess.run(
            [
                CARGO,
                "metadata",
                "--quiet",
                "--format-version",
                "1",
                "--offline",
                "--no-deps",
                "--manifest-path",
                path,
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        try:
            ret.check_returncode()
            stdout = ret.stdout.decode()
            stderr = ret.stderr.decode()

            if stderr:
                print(stderr, file=sys.stderr)

            return Metadata.from_json(stdout)

        except subprocess.CalledProcessError as exc:
            if exc.stdout:
                print(exc.stdout.decode(), file=sys.stdout)
            if exc.stderr:
                print(exc.stderr.decode(), file=sys.stderr)
            raise

    @property
    def packages(self) -> list[Package]:
        return [Package(package) for package in self._data["packages"]]

    @property
    def target_directory(self) -> str:
        return self._data["target_directory"]

    def is_workspace(self) -> bool:
        """
        Returns `True` if the metadata looks like a cargo workspace.

        A workspace with a single member is equivalent to no workspace (since
        no path dependencies are possible with only one workspace member), so
        the naive check for "at least two workspace members" is enough.
        """

        return len(self.packages) >= 2

    def is_bin(self) -> bool:
        """
        Returns `True` if there are any "proper" binary targets.
        """

        for package in self.packages:
            if package.is_bin():
                return True
        return False

    def is_lib(self) -> bool:
        """
        Returns `True` if the crate has a `lib` or `proc-macro` build target.
        Always returns `False` for cargo workspaces.
        """

        packages = self.packages

        # do not report libs from workspaces until this is actually supported
        if len(packages) >= 2:
            return False

        for package in packages:
            if package.is_lib():
                return True
        return False

    def is_cdylib(self) -> bool:
        """
        Returns `True` if there are any `cdylib` (C shared library) targets.
        """

        for package in self.packages:
            if package.is_cdylib():
                return True
        return False

    def get_binaries(self) -> set[str]:
        """
        Returns the union of all sets of "proper" binary build targets of all
        crates in this workspace (or of the single crate for non-workspace
        crates).
        """

        bins = set()
        for package in self.packages:
            bins.update(package.get_binaries())
        return bins

    def get_feature_flags_for_workspace_members(self, flags: FeatureFlags) -> dict[str, FeatureFlags]:
        """
        Resolves the transitive closure of enabled features for intra-workspace
        crate dependencies, taking `required-features` for binary targets and
        passed feature flags into account.
        """

        members = {package.name for package in self.packages}

        # keep track of workspace members and which features are enabled for them:
        # - determine union of enabled features for all workspace members
        member_features: dict[str, set[str]] = dict()
        # - determine whether default features are enabled for workspace members
        member_defaults: dict[str, bool] = dict()

        # collect required features of "bin" and "cdylib" targets
        required_features: dict[str, set[str]] = dict()

        # apply feature flags to all packages
        for package in self.packages:
            if flags.all_features:
                features = member_features.get(package.name) or set()
                features.update(package.get_feature_names())
                member_features[package.name] = features

            if flags.features:
                features = member_features.get(package.name) or set()
                for feature in flags.features:
                    if feature in package.get_feature_names():
                        features.add(feature)
                member_features[package.name] = features

            if flags.no_default_features:
                member_defaults[package.name] = False

            # ensure that the mapping includes data for all packages
            features = member_features.get(package.name) or set()
            member_features[package.name] = features

        # enable required and default features of binary targets
        for package in self.packages:
            for target in package.targets:
                if ("bin" in target.kind and "bin" in target.crate_types) or ("cdylib" in target.kind and "cdylib" in target.crate_types):
                    if reqs := target.required_features:
                        if package.name not in required_features.keys():
                            required_features[package.name] = set(reqs)
                        else:
                            required_features[package.name].update(reqs)
                    if not flags.no_default_features:
                        member_defaults[package.name] = True

        for package in self.packages:
            for dep in filter(lambda pkg: pkg.path is not None, package.dependencies):
                features = member_features.get(dep.name) or set()
                features.update(dep.features)
                member_features[dep.name] = features

                defaults = member_defaults.get(dep.name) or False
                defaults = defaults or dep.uses_default_features
                member_defaults[dep.name] = defaults

        # unconditionally add required features to resolved features:
        # "cargo build" skips building targets for which required features are
        # not enabled, but for package builds this behaviour doesn't make sense
        for name, required in required_features.items():
            member_features[name].update(required)

        # turn on default features for all workspace members that are not
        # explicitly referenced by other workspace members
        for package in self.packages:
            if package.name not in member_defaults.keys():
                member_defaults[package.name] = True

        # enable features pulled in by feature dependencies of enabled features
        for package in self.packages:
            deps_real_names = {dep.rename or dep.name: dep.name for dep in package.dependencies}

            enabled = member_features[package.name].copy()
            if member_defaults[package.name] and "default" in package.get_feature_names():
                enabled.add("default")

            while True:
                new: set[str] = set()

                for feature_name, deps in package.features.items():
                    if feature_name not in enabled:
                        continue

                    for fdep in deps:
                        # named optional dependency
                        if fdep.startswith("dep:"):
                            # cargo builds all workspace members even if they are optional
                            continue

                        # dependency/feature
                        if "/" in fdep and "?/" not in fdep:
                            name, feat = fdep.split("/")

                            if deps_real_names[name] in members:
                                real_name = deps_real_names[name]
                                member_features[real_name].add(feat)
                            continue

                        # dependency?/feature
                        if "?/" in fdep:
                            name, feat = fdep.split("?/")

                            if deps_real_names[name] in members:
                                real_name = deps_real_names[name]
                                member_features[real_name].add(feat)
                            continue

                        # feature name
                        if fdep not in enabled:
                            new.add(fdep)
                            continue

                # continue until set of enabled "proper" features no longer changes
                if new:
                    enabled.update(new)
                else:
                    break

        # construct feature flags from collected settings
        member_flags: dict[str, FeatureFlags] = dict()
        for package in self.packages:
            if member_features[package.name] == package.get_feature_names() and flags.all_features:
                flag_no_default_features = False
                flag_all_features = True
                flag_features = list()
            else:
                flag_no_default_features = not member_defaults[package.name]
                flag_all_features = False
                flag_features = list(sorted(member_features[package.name]))

            flags = FeatureFlags(all_features=flag_all_features, no_default_features=flag_no_default_features, features=flag_features)
            member_flags[package.name] = flags

        return member_flags



import importlib.resources

from cargo2rpm.metadata import Metadata


def load_metadata_from_resource(filename: str) -> Metadata:
    """
    Utility function for locating and loading test data (i.e. the JSON dump from
    `cargo metadata`) as identified by its file name, and parse it into a
    `Metadata` object.
    """

    data = importlib.resources.files("cargo2rpm.testdata").joinpath(filename).read_text()
    return Metadata.from_json(data)


def short_repr(obj) -> str:
    """
    Utility function for returning a truncated `repr` of the object that was
    passed as an argument. Used for identifying test cases in parametrized
    `pytest` tests.
    """

    s = repr(obj)
    if len(s) >= 22:
        return s[0:22] + ".."
    else:
        return s



from cargo2rpm.semver import Comparator, Op, Version
from cargo2rpm.utils import short_repr

import pytest


@pytest.mark.parametrize(
    "string,expected",
    [
        ("1", Comparator(Op.CARET, 1, None, None, None)),
        ("1.2", Comparator(Op.CARET, 1, 2, None, None)),
        ("1.2.3", Comparator(Op.CARET, 1, 2, 3, None)),
        ("1.2.3-alpha.1", Comparator(Op.CARET, 1, 2, 3, "alpha.1")),
        ("1.2.3-alpha2", Comparator(Op.CARET, 1, 2, 3, "alpha2")),
        ("=0", Comparator(Op.EXACT, 0, None, None, None)),
        ("=0.11", Comparator(Op.EXACT, 0, 11, None, None)),
        ("=0.1.17", Comparator(Op.EXACT, 0, 1, 17, None)),
        ("=0.2.37-alpha.1", Comparator(Op.EXACT, 0, 2, 37, "alpha.1")),
        ("=0.2.37-alpha2", Comparator(Op.EXACT, 0, 2, 37, "alpha2")),
        (">1", Comparator(Op.GREATER, 1, None, None, None)),
        (">1.2", Comparator(Op.GREATER, 1, 2, None, None)),
        (">1.2.3", Comparator(Op.GREATER, 1, 2, 3, None)),
        (">1.2.3-alpha.1", Comparator(Op.GREATER, 1, 2, 3, "alpha.1")),
        (">1.2.3-alpha2", Comparator(Op.GREATER, 1, 2, 3, "alpha2")),
        (">=0", Comparator(Op.GREATER_EQ, 0, None, None, None)),
        (">=0.11", Comparator(Op.GREATER_EQ, 0, 11, None, None)),
        (">=0.1.17", Comparator(Op.GREATER_EQ, 0, 1, 17, None)),
        (">=0.2.37-alpha.1", Comparator(Op.GREATER_EQ, 0, 2, 37, "alpha.1")),
        (">=0.2.37-alpha2", Comparator(Op.GREATER_EQ, 0, 2, 37, "alpha2")),
        ("<1", Comparator(Op.LESS, 1, None, None, None)),
        ("<1.2", Comparator(Op.LESS, 1, 2, None, None)),
        ("<1.2.3", Comparator(Op.LESS, 1, 2, 3, None)),
        ("<1.2.3-alpha.1", Comparator(Op.LESS, 1, 2, 3, "alpha.1")),
        ("<1.2.3-alpha2", Comparator(Op.LESS, 1, 2, 3, "alpha2")),
        ("<=0", Comparator(Op.LESS_EQ, 0, None, None, None)),
        ("<=0.11", Comparator(Op.LESS_EQ, 0, 11, None, None)),
        ("<=0.1.17", Comparator(Op.LESS_EQ, 0, 1, 17, None)),
        ("<=0.2.37-alpha.1", Comparator(Op.LESS_EQ, 0, 2, 37, "alpha.1")),
        ("<=0.2.37-alpha2", Comparator(Op.LESS_EQ, 0, 2, 37, "alpha2")),
        ("~1", Comparator(Op.TILDE, 1, None, None, None)),
        ("~1.2", Comparator(Op.TILDE, 1, 2, None, None)),
        ("~1.2.3", Comparator(Op.TILDE, 1, 2, 3, None)),
        ("~1.2.3-alpha.1", Comparator(Op.TILDE, 1, 2, 3, "alpha.1")),
        ("~1.2.3-alpha2", Comparator(Op.TILDE, 1, 2, 3, "alpha2")),
        ("^0", Comparator(Op.CARET, 0, None, None, None)),
        ("^0.11", Comparator(Op.CARET, 0, 11, None, None)),
        ("^0.1.17", Comparator(Op.CARET, 0, 1, 17, None)),
        ("^0.2.37-alpha.1", Comparator(Op.CARET, 0, 2, 37, "alpha.1")),
        ("^0.2.37-alpha2", Comparator(Op.CARET, 0, 2, 37, "alpha2")),
        ("1.*", Comparator(Op.WILDCARD, 1, None, None, None)),
        ("1.*.*", Comparator(Op.WILDCARD, 1, None, None, None)),
        ("1.2.*", Comparator(Op.WILDCARD, 1, 2, None, None)),
    ],
    ids=short_repr,
)
def test_parse_comparator(string: str, expected: Comparator):
    assert Comparator.parse(string) == expected


@pytest.mark.parametrize(
    "string,expected_exc,expected_err",
    [
        ("foo", ValueError, "Invalid version requirement"),
        ("0.*.0", ValueError, "Invalid wildcard requirement"),
        ("0-alpha.1", ValueError, "Invalid pre-release requirement"),
        ("01.2.3", ValueError, "Invalid version requirement"),
        ("1.02.3", ValueError, "Invalid version requirement"),
        ("1.2.03", ValueError, "Invalid version requirement"),
    ],
    ids=short_repr,
)
def test_parse_comparator_fail(string: str, expected_exc, expected_err: str):
    with pytest.raises(expected_exc) as exc:
        Comparator.parse(string)
    assert expected_err in str(exc.value)


@pytest.mark.parametrize(
    "comparator,version,expected",
    [
        # =
        (Comparator(Op.EXACT, 1, 2, 3, None), Version(1, 2, 3), True),
        (Comparator(Op.EXACT, 1, 2, 3, None), Version(0, 1, 2), False),
        # >
        (Comparator(Op.GREATER, 1, 2, 3, None), Version(1, 2, 4), True),
        (Comparator(Op.GREATER, 1, 2, 3, None), Version(1, 2, 2), False),
        # >=
        (Comparator(Op.GREATER_EQ, 1, 2, 3, None), Version(1, 2, 3), True),
        (Comparator(Op.GREATER_EQ, 1, 2, 3, None), Version(1, 2, 4), True),
        (Comparator(Op.GREATER_EQ, 1, 2, 3, None), Version(1, 2, 2), False),
        # <
        (Comparator(Op.LESS, 1, 2, 3, None), Version(1, 2, 2), True),
        (Comparator(Op.LESS, 1, 2, 3, None), Version(1, 2, 4), False),
        # <=
        (Comparator(Op.LESS_EQ, 1, 2, 3, None), Version(1, 2, 3), True),
        (Comparator(Op.LESS_EQ, 1, 2, 3, None), Version(1, 2, 2), True),
        (Comparator(Op.LESS_EQ, 1, 2, 3, None), Version(1, 2, 4), False),
        # ~
        (Comparator(Op.TILDE, 1, 2, 3, None), Version(1, 2, 3), True),
        (Comparator(Op.TILDE, 1, 2, 3, None), Version(1, 2, 4), True),
        (Comparator(Op.TILDE, 1, 2, 3, None), Version(1, 2, 2), False),
        (Comparator(Op.TILDE, 1, 2, 3, None), Version(1, 3, 0), False),
        # ^
        (Comparator(Op.CARET, 1, 2, 3, None), Version(1, 2, 3), True),
        (Comparator(Op.CARET, 1, 2, 3, None), Version(1, 2, 4), True),
        (Comparator(Op.CARET, 1, 2, 3, None), Version(1, 3, 0), True),
        (Comparator(Op.CARET, 1, 2, 3, None), Version(1, 2, 2), False),
        (Comparator(Op.CARET, 1, 2, 3, None), Version(2, 0, 0), False),
        # *
        (Comparator(Op.WILDCARD, 1, None, None, None), Version(1, 0, 0), True),
        (Comparator(Op.WILDCARD, 1, None, None, None), Version(2, 0, 0), False),
        (Comparator(Op.WILDCARD, 1, 2, None, None), Version(1, 2, 0), True),
        (Comparator(Op.WILDCARD, 1, 2, None, None), Version(1, 1, 0), False),
    ],
    ids=repr,
)
def test_comparator_contains(comparator: Comparator, version: Version, expected: bool):
    assert (version in comparator) == expected



from cargo2rpm.rpm import _provides_crate, _provides_feature
from cargo2rpm.utils import load_metadata_from_resource, short_repr

import pytest


@pytest.mark.parametrize(
    "filename,expected",
    [
        ("ahash-0.8.3.json", "crate(ahash) = 0.8.3"),
        ("assert_cmd-2.0.8.json", "crate(assert_cmd) = 2.0.8"),
        ("assert_fs-1.0.10.json", "crate(assert_fs) = 1.0.10"),
        ("autocfg-1.1.0.json", "crate(autocfg) = 1.1.0"),
        ("bstr-1.2.0.json", "crate(bstr) = 1.2.0"),
        ("cfg-if-1.0.0.json", "crate(cfg-if) = 1.0.0"),
        ("clap-4.1.4.json", "crate(clap) = 4.1.4"),
        ("gstreamer-0.19.7.json", "crate(gstreamer) = 0.19.7"),
        ("human-panic-1.1.0.json", "crate(human-panic) = 1.1.0"),
        ("libc-0.2.139.json", "crate(libc) = 0.2.139"),
        ("predicates-2.1.5.json", "crate(predicates) = 2.1.5"),
        ("proc-macro2-1.0.50.json", "crate(proc-macro2) = 1.0.50"),
        ("quote-1.0.23.json", "crate(quote) = 1.0.23"),
        ("rand-0.8.5.json", "crate(rand) = 0.8.5"),
        ("rand_core-0.6.4.json", "crate(rand_core) = 0.6.4"),
        ("rust_decimal-1.28.0.json", "crate(rust_decimal) = 1.28.0"),
        ("rustix-0.36.8.json", "crate(rustix) = 0.36.8"),
        ("serde-1.0.152.json", "crate(serde) = 1.0.152"),
        ("serde_derive-1.0.152.json", "crate(serde_derive) = 1.0.152"),
        ("sha1collisiondetection-0.3.1.json", "crate(sha1collisiondetection) = 0.3.1"),
        ("syn-1.0.107.json", "crate(syn) = 1.0.107"),
        ("time-0.3.17.json", "crate(time) = 0.3.17"),
        ("tokio-1.25.0.json", "crate(tokio) = 1.25.0"),
        ("unicode-xid-0.2.4.json", "crate(unicode-xid) = 0.2.4"),
        ("zbus-3.8.0.json", "crate(zbus) = 3.8.0"),
    ],
    ids=short_repr,
)
def test_provides_crate(filename: str, expected: str):
    metadata = load_metadata_from_resource(filename)
    assert _provides_crate(metadata.packages[0]) == expected


@pytest.mark.parametrize(
    "filename,expected",
    [
        ("ahash-0.8.3.json", "crate(ahash/default) = 0.8.3"),
        ("assert_cmd-2.0.8.json", "crate(assert_cmd/default) = 2.0.8"),
        ("assert_fs-1.0.10.json", "crate(assert_fs/default) = 1.0.10"),
        ("autocfg-1.1.0.json", "crate(autocfg/default) = 1.1.0"),
        ("bstr-1.2.0.json", "crate(bstr/default) = 1.2.0"),
        ("cfg-if-1.0.0.json", "crate(cfg-if/default) = 1.0.0"),
        ("clap-4.1.4.json", "crate(clap/default) = 4.1.4"),
        ("gstreamer-0.19.7.json", "crate(gstreamer/default) = 0.19.7"),
        ("human-panic-1.1.0.json", "crate(human-panic/default) = 1.1.0"),
        ("libc-0.2.139.json", "crate(libc/default) = 0.2.139"),
        ("predicates-2.1.5.json", "crate(predicates/default) = 2.1.5"),
        ("proc-macro2-1.0.50.json", "crate(proc-macro2/default) = 1.0.50"),
        ("quote-1.0.23.json", "crate(quote/default) = 1.0.23"),
        ("rand-0.8.5.json", "crate(rand/default) = 0.8.5"),
        ("rand_core-0.6.4.json", "crate(rand_core/default) = 0.6.4"),
        ("rust_decimal-1.28.0.json", "crate(rust_decimal/default) = 1.28.0"),
        ("rustix-0.36.8.json", "crate(rustix/default) = 0.36.8"),
        ("serde-1.0.152.json", "crate(serde/default) = 1.0.152"),
        ("serde_derive-1.0.152.json", "crate(serde_derive/default) = 1.0.152"),
        ("sha1collisiondetection-0.3.1.json", "crate(sha1collisiondetection/default) = 0.3.1"),
        ("syn-1.0.107.json", "crate(syn/default) = 1.0.107"),
        ("time-0.3.17.json", "crate(time/default) = 0.3.17"),
        ("tokio-1.25.0.json", "crate(tokio/default) = 1.25.0"),
        ("unicode-xid-0.2.4.json", "crate(unicode-xid/default) = 0.2.4"),
        ("zbus-3.8.0.json", "crate(zbus/default) = 3.8.0"),
    ],
    ids=short_repr,
)
def test_provides_feature(filename: str, expected: str):
    metadata = load_metadata_from_resource(filename)
    assert _provides_feature(metadata.packages[0], "default") == expected



from typing import Optional

from cargo2rpm.semver import Comparator, Op, Version, VersionReq
from cargo2rpm.utils import short_repr

import pytest


@pytest.mark.parametrize(
    "string,expected",
    [
        ("^0.1", VersionReq([Comparator(Op.CARET, 0, 1, None, None)])),
        ("~0.1", VersionReq([Comparator(Op.TILDE, 0, 1, None, None)])),
        ("^1.2", VersionReq([Comparator(Op.CARET, 1, 2, None, None)])),
        ("~1.2", VersionReq([Comparator(Op.TILDE, 1, 2, None, None)])),
        (">=1.2.3, <2.0.0", VersionReq([Comparator(Op.GREATER_EQ, 1, 2, 3, None), Comparator(Op.LESS, 2, 0, 0, None)])),
        (">=0.1.7, <0.2.0", VersionReq([Comparator(Op.GREATER_EQ, 0, 1, 7, None), Comparator(Op.LESS, 0, 2, 0, None)])),
        ("*", VersionReq([])),
    ],
    ids=short_repr,
)
def test_parse_version_req(string: str, expected: VersionReq):
    assert VersionReq.parse(string) == expected


def test_parse_version_req_fail():
    with pytest.raises(ValueError) as exc:
        VersionReq.parse("")
    assert "Invalid version requirement (empty string)" in str(exc.value)


@pytest.mark.parametrize(
    "string,expected",
    [
        ("=1.2.3", [Comparator(Op.EXACT, 1, 2, 3, None)]),
        ("=1.2", [Comparator(Op.GREATER_EQ, 1, 2, 0, None), Comparator(Op.LESS, 1, 3, 0, None)]),
        ("=1", [Comparator(Op.GREATER_EQ, 1, 0, 0, None), Comparator(Op.LESS, 2, 0, 0, None)]),
        (">1.2.3", [Comparator(Op.GREATER, 1, 2, 3, None)]),
        (">1.2", [Comparator(Op.GREATER_EQ, 1, 3, 0, None)]),
        (">1", [Comparator(Op.GREATER_EQ, 2, 0, 0, None)]),
        (">=1.2.3", [Comparator(Op.GREATER_EQ, 1, 2, 3, None)]),
        (">=1.2", [Comparator(Op.GREATER_EQ, 1, 2, 0, None)]),
        (">=1", [Comparator(Op.GREATER_EQ, 1, 0, 0, None)]),
        ("<1.2.3", [Comparator(Op.LESS, 1, 2, 3, None)]),
        ("<1.2", [Comparator(Op.LESS, 1, 2, 0, None)]),
        ("<1", [Comparator(Op.LESS, 1, 0, 0, None)]),
        ("<=1.2.3", [Comparator(Op.LESS_EQ, 1, 2, 3, None)]),
        ("<=1.2", [Comparator(Op.LESS, 1, 3, 0, None)]),
        ("<=1", [Comparator(Op.LESS, 2, 0, 0, None)]),
        ("~1.2.3", [Comparator(Op.GREATER_EQ, 1, 2, 3, None), Comparator(Op.LESS, 1, 3, 0, None)]),
        ("~1.2", [Comparator(Op.GREATER_EQ, 1, 2, 0, None), Comparator(Op.LESS, 1, 3, 0, None)]),
        ("~1", [Comparator(Op.GREATER_EQ, 1, 0, 0, None), Comparator(Op.LESS, 2, 0, 0, None)]),
        ("^1.2.3", [Comparator(Op.GREATER_EQ, 1, 2, 3, None), Comparator(Op.LESS, 2, 0, 0, None)]),
        ("^0.1.2", [Comparator(Op.GREATER_EQ, 0, 1, 2, None), Comparator(Op.LESS, 0, 2, 0, None)]),
        ("^0.0.1", [Comparator(Op.EXACT, 0, 0, 1, None)]),
        ("^1.2", [Comparator(Op.GREATER_EQ, 1, 2, 0, None), Comparator(Op.LESS, 2, 0, 0, None)]),
        ("^0.1", [Comparator(Op.GREATER_EQ, 0, 1, 0, None), Comparator(Op.LESS, 0, 2, 0, None)]),
        ("^0.0", [Comparator(Op.GREATER_EQ, 0, 0, 0, None), Comparator(Op.LESS, 0, 1, 0, None)]),
        ("^1", [Comparator(Op.GREATER_EQ, 1, 0, 0, None), Comparator(Op.LESS, 2, 0, 0, None)]),
        ("1.2.*", [Comparator(Op.GREATER_EQ, 1, 2, 0, None), Comparator(Op.LESS, 1, 3, 0, None)]),
        ("1.*", [Comparator(Op.GREATER_EQ, 1, 0, 0, None), Comparator(Op.LESS, 2, 0, 0, None)]),
        ("1.*.*", [Comparator(Op.GREATER_EQ, 1, 0, 0, None), Comparator(Op.LESS, 2, 0, 0, None)]),
    ],
    ids=short_repr,
)
def test_normalize_comparator(string: str, expected: list[Comparator]):
    assert Comparator.parse(string).normalize() == expected


@pytest.mark.parametrize(
    "comparator,feature,expected",
    [
        ("=1.2.3", None, "crate(foo) = 1.2.3"),
        ("=1.2.3", "default", "crate(foo/default) = 1.2.3"),
        (">1.2.3", None, "crate(foo) > 1.2.3"),
        (">1.2.3", "default", "crate(foo/default) > 1.2.3"),
        (">=1.2.3", None, "crate(foo) >= 1.2.3"),
        (">=1.2.3", "default", "crate(foo/default) >= 1.2.3"),
        ("<1.2.3", None, "crate(foo) < 1.2.3~"),
        ("<1.2.3", "default", "crate(foo/default) < 1.2.3~"),
        ("<=1.2.3", None, "crate(foo) <= 1.2.3"),
        ("<=1.2.3", "default", "crate(foo/default) <= 1.2.3"),
    ],
    ids=short_repr,
)
def test_comparator_to_rpm(comparator: str, feature: Optional[str], expected: str):
    assert Comparator.parse(comparator).to_rpm("foo", feature) == expected


@pytest.mark.parametrize(
    "req,feature,expected",
    [
        ("*", None, "crate(foo)"),
        ("*", "default", "crate(foo/default)"),
        ("=1.2.3", None, "crate(foo) = 1.2.3"),
        ("=1.2.3", "default", "crate(foo/default) = 1.2.3"),
        ("=1.2", None, "(crate(foo) >= 1.2.0 with crate(foo) < 1.3.0~)"),
        ("=1.2", "default", "(crate(foo/default) >= 1.2.0 with crate(foo/default) < 1.3.0~)"),
        ("=1", None, "(crate(foo) >= 1.0.0 with crate(foo) < 2.0.0~)"),
        ("=1", "default", "(crate(foo/default) >= 1.0.0 with crate(foo/default) < 2.0.0~)"),
        (">1.2.3", None, "crate(foo) > 1.2.3"),
        (">1.2.3", "default", "crate(foo/default) > 1.2.3"),
        (">1.2", None, "crate(foo) >= 1.3.0"),
        (">1.2", "default", "crate(foo/default) >= 1.3.0"),
        (">1", None, "crate(foo) >= 2.0.0"),
        (">1", "default", "crate(foo/default) >= 2.0.0"),
        (">=1.2.3", None, "crate(foo) >= 1.2.3"),
        (">=1.2.3", "default", "crate(foo/default) >= 1.2.3"),
        (">=1.2", None, "crate(foo) >= 1.2.0"),
        (">=1.2", "default", "crate(foo/default) >= 1.2.0"),
        (">=1", None, "crate(foo) >= 1.0.0"),
        (">=1", "default", "crate(foo/default) >= 1.0.0"),
        ("<1.2.3", None, "crate(foo) < 1.2.3~"),
        ("<1.2.3", "default", "crate(foo/default) < 1.2.3~"),
        ("<1.2", None, "crate(foo) < 1.2.0~"),
        ("<1.2", "default", "crate(foo/default) < 1.2.0~"),
        ("<1", None, "crate(foo) < 1.0.0~"),
        ("<1", "default", "crate(foo/default) < 1.0.0~"),
        ("<=1.2.3", None, "crate(foo) <= 1.2.3"),
        ("<=1.2.3", "default", "crate(foo/default) <= 1.2.3"),
        ("<=1.2", None, "crate(foo) < 1.3.0~"),
        ("<=1.2", "default", "crate(foo/default) < 1.3.0~"),
        ("<=1", None, "crate(foo) < 2.0.0~"),
        ("<=1", "default", "crate(foo/default) < 2.0.0~"),
        ("~1.2.3", None, "(crate(foo) >= 1.2.3 with crate(foo) < 1.3.0~)"),
        ("~1.2.3", "default", "(crate(foo/default) >= 1.2.3 with crate(foo/default) < 1.3.0~)"),
        ("~1.2", None, "(crate(foo) >= 1.2.0 with crate(foo) < 1.3.0~)"),
        ("~1.2", "default", "(crate(foo/default) >= 1.2.0 with crate(foo/default) < 1.3.0~)"),
        ("~1", None, "(crate(foo) >= 1.0.0 with crate(foo) < 2.0.0~)"),
        ("~1", "default", "(crate(foo/default) >= 1.0.0 with crate(foo/default) < 2.0.0~)"),
        ("^1.2.3", None, "(crate(foo) >= 1.2.3 with crate(foo) < 2.0.0~)"),
        ("^1.2.3", "default", "(crate(foo/default) >= 1.2.3 with crate(foo/default) < 2.0.0~)"),
        ("^0.1.2", None, "(crate(foo) >= 0.1.2 with crate(foo) < 0.2.0~)"),
        ("^0.1.2", "default", "(crate(foo/default) >= 0.1.2 with crate(foo/default) < 0.2.0~)"),
        ("^0.0.1", None, "crate(foo) = 0.0.1"),
        ("^0.0.1", "default", "crate(foo/default) = 0.0.1"),
        ("^1.2", None, "(crate(foo) >= 1.2.0 with crate(foo) < 2.0.0~)"),
        ("^1.2", "default", "(crate(foo/default) >= 1.2.0 with crate(foo/default) < 2.0.0~)"),
        ("^0.1", None, "(crate(foo) >= 0.1.0 with crate(foo) < 0.2.0~)"),
        ("^0.1", "default", "(crate(foo/default) >= 0.1.0 with crate(foo/default) < 0.2.0~)"),
        ("^0.0", None, "(crate(foo) >= 0.0.0 with crate(foo) < 0.1.0~)"),
        ("^0.0", "default", "(crate(foo/default) >= 0.0.0 with crate(foo/default) < 0.1.0~)"),
        ("^1", None, "(crate(foo) >= 1.0.0 with crate(foo) < 2.0.0~)"),
        ("^1", "default", "(crate(foo/default) >= 1.0.0 with crate(foo/default) < 2.0.0~)"),
        ("1.2.3", None, "(crate(foo) >= 1.2.3 with crate(foo) < 2.0.0~)"),
        ("1.2.3", "default", "(crate(foo/default) >= 1.2.3 with crate(foo/default) < 2.0.0~)"),
        ("0.1.2", None, "(crate(foo) >= 0.1.2 with crate(foo) < 0.2.0~)"),
        ("0.1.2", "default", "(crate(foo/default) >= 0.1.2 with crate(foo/default) < 0.2.0~)"),
        ("0.0.1", None, "crate(foo) = 0.0.1"),
        ("0.0.1", "default", "crate(foo/default) = 0.0.1"),
        ("1.2", None, "(crate(foo) >= 1.2.0 with crate(foo) < 2.0.0~)"),
        ("1.2", "default", "(crate(foo/default) >= 1.2.0 with crate(foo/default) < 2.0.0~)"),
        ("0.1", None, "(crate(foo) >= 0.1.0 with crate(foo) < 0.2.0~)"),
        ("0.1", "default", "(crate(foo/default) >= 0.1.0 with crate(foo/default) < 0.2.0~)"),
        ("0.0", None, "(crate(foo) >= 0.0.0 with crate(foo) < 0.1.0~)"),
        ("0.0", "default", "(crate(foo/default) >= 0.0.0 with crate(foo/default) < 0.1.0~)"),
        ("1", None, "(crate(foo) >= 1.0.0 with crate(foo) < 2.0.0~)"),
        ("1", "default", "(crate(foo/default) >= 1.0.0 with crate(foo/default) < 2.0.0~)"),
        ("1.2.*", None, "(crate(foo) >= 1.2.0 with crate(foo) < 1.3.0~)"),
        ("1.2.*", "default", "(crate(foo/default) >= 1.2.0 with crate(foo/default) < 1.3.0~)"),
        ("1.*", None, "(crate(foo) >= 1.0.0 with crate(foo) < 2.0.0~)"),
        ("1.*", "default", "(crate(foo/default) >= 1.0.0 with crate(foo/default) < 2.0.0~)"),
        ("1.*.*", None, "(crate(foo) >= 1.0.0 with crate(foo) < 2.0.0~)"),
        ("1.*.*", "default", "(crate(foo/default) >= 1.0.0 with crate(foo/default) < 2.0.0~)"),
        # some real-world test cases with pre-releases
        ("^0.1.0-alpha.4", "default", "(crate(foo/default) >= 0.1.0~alpha.4 with crate(foo/default) < 0.2.0~)"),
        ("=1.0.0-alpha.5", "default", "crate(foo/default) = 1.0.0~alpha.5"),
    ],
    ids=short_repr,
)
def test_version_req_to_rpm(req: str, feature: Optional[str], expected: str):
    assert VersionReq.parse(req).to_rpm("foo", feature) == expected


def test_version_req_to_rpm_fail():
    with pytest.raises(ValueError) as exc:
        VersionReq.parse(">=0.1.0, <0.2.0, >=0.3.0, <1.0.0").to_rpm("foo", None)
    assert "Using more than 2 comparators is not supported by RPM." in str(exc.value)


@pytest.mark.parametrize(
    "req,version,expected",
    [
        # =
        (VersionReq.parse("=0.1.2"), Version(0, 1, 2), True),
        (VersionReq.parse("=0.1.2"), Version(0, 1, 1), False),
        (VersionReq.parse("=0.1.2"), Version(0, 1, 3), False),
        # >=, <
        (VersionReq.parse(">=0.13, <0.22"), Version(0, 13, 0), True),
        (VersionReq.parse(">=0.13, <0.22"), Version(0, 21, 42), True),
        (VersionReq.parse(">=0.13, <0.22"), Version(0, 12, 0), False),
        (VersionReq.parse(">=0.13, <0.22"), Version(0, 22, 0), False),
        # ~
        (VersionReq.parse("~1.2.3"), Version(1, 2, 3), True),
        (VersionReq.parse("~1.2.3"), Version(1, 2, 2), False),
        (VersionReq.parse("~1.2.3"), Version(1, 3, 0), False),
        (VersionReq.parse("~1.2.3"), Version(2, 0, 0), False),
        # ^
        (VersionReq.parse("^1.2.3"), Version(1, 2, 3), True),
        (VersionReq.parse("^1.2.3"), Version(1, 2, 2), False),
        (VersionReq.parse("^1.2.3"), Version(2, 0, 0), False),
    ],
    ids=repr,
)
def test_version_req_contains(req: VersionReq, version: Version, expected: bool):
    print(req.comparators)
    print(req.normalize().comparators)
    assert (version in req) == expected



import textwrap

from cargo2rpm.utils import load_metadata_from_resource, short_repr

import pytest


@pytest.mark.parametrize(
    "filename,expected",
    [
        ("ahash-0.8.3.json", "A non-cryptographic hash function using AES-NI for high performance."),
        ("aho-corasick-1.0.2.json", "Fast multiple substring searching."),
        ("assert_cmd-2.0.8.json", "Test CLI Applications."),
        ("assert_fs-1.0.10.json", "Filesystem fixtures and assertions for testing."),
        ("autocfg-1.1.0.json", "Automatic cfg for Rust compiler features."),
        ("bstr-1.2.0.json", "A string type that is not required to be valid UTF-8."),
        (
            "cfg-if-1.0.0.json",
            textwrap.dedent(
                """\
                A macro to ergonomically define an item depending on a large number of
                #[cfg] parameters. Structured like an if-else chain, the first matching
                branch is the item that gets emitted."""
            ),
        ),
        (
            "clap-4.1.4.json",
            textwrap.dedent(
                """\
                A simple to use, efficient, and full-featured Command Line Argument
                Parser."""
            ),
        ),
        ("gstreamer-0.19.7.json", "Rust bindings for GStreamer."),
        ("human-panic-1.1.0.json", "Panic messages for humans."),
        ("hyperfine-1.15.0.json", "A command-line benchmarking tool."),
        ("iri-string-0.7.0.json", "IRI as string types."),
        ("libc-0.2.139.json", "Raw FFI bindings to platform libraries like libc."),
        ("predicates-2.1.5.json", "An implementation of boolean-valued predicate functions."),
        (
            "proc-macro2-1.0.50.json",
            textwrap.dedent(
                """\
                A substitute implementation of the compiler's `proc_macro` API to
                decouple token-based libraries from the procedural macro use case."""
            ),
        ),
        ("quote-1.0.23.json", "Quasi-quoting macro quote!(...)."),
        ("rand-0.8.5.json", "Random number generators and other randomness functionality."),
        ("rand_core-0.6.4.json", "Core random number generator traits and tools for implementation."),
        (
            "regex-1.8.4.json",
            textwrap.dedent(
                """\
                An implementation of regular expressions for Rust. This implementation
                uses finite automata and guarantees linear time matching on all inputs."""
            ),
        ),
        ("regex-syntax-0.7.2.json", "A regular expression parser."),
        ("rpm-sequoia-1.2.0.json", "An implementation of the RPM PGP interface using Sequoia."),
        (
            "rust_decimal-1.28.0.json",
            textwrap.dedent(
                """\
                Decimal number implementation written in pure Rust suitable for
                financial and fixed-precision calculations."""
            ),
        ),
        ("rustix-0.36.8.json", "Safe Rust bindings to POSIX/Unix/Linux/Winsock2-like syscalls."),
        ("serde-1.0.152.json", "A generic serialization/deserialization framework."),
        ("serde_derive-1.0.152.json", "Macros 1.1 implementation of #[derive(Serialize, Deserialize)]."),
        ("sha1collisiondetection-0.3.1.json", "SHA-1 hash function with collision detection and mitigation."),
        ("syn-1.0.107.json", "Parser for Rust source code."),
        (
            "time-0.3.17.json",
            textwrap.dedent(
                """\
                Date and time library. Fully interoperable with the standard library.
                Mostly compatible with #![no_std]."""
            ),
        ),
        (
            "tokio-1.25.0.json",
            textwrap.dedent(
                """\
                An event-driven, non-blocking I/O platform for writing asynchronous I/O
                backed applications."""
            ),
        ),
        (
            "unicode-xid-0.2.4.json",
            textwrap.dedent(
                """\
                Determine whether characters have the XID_Start or XID_Continue
                properties according to Unicode Standard Annex #31."""
            ),
        ),
        ("zbus-3.8.0.json", "API for D-Bus communication."),
        ("zoxide-0.9.0.json", "A smarter cd command for your terminal."),
    ],
    ids=short_repr,
)
def test_metadata_get_description(filename: str, expected: str):
    metadata = load_metadata_from_resource(filename)
    assert metadata.packages[0].get_description() == expected


@pytest.mark.parametrize(
    "filename,expected",
    [
        ("ahash-0.8.3.json", "Non-cryptographic hash function using AES-NI for high performance"),
        ("aho-corasick-1.0.2.json", "Fast multiple substring searching"),
        ("assert_cmd-2.0.8.json", "Test CLI Applications"),
        ("assert_fs-1.0.10.json", "Filesystem fixtures and assertions for testing"),
        ("autocfg-1.1.0.json", "Automatic cfg for Rust compiler features"),
        ("bstr-1.2.0.json", "String type that is not required to be valid UTF-8"),
        ("cfg-if-1.0.0.json", "Macro to ergonomically define an item depending on a large number of #[cfg] parameters"),
        ("clap-4.1.4.json", "Simple to use, efficient, and full-featured Command Line Argument Parser"),
        ("gstreamer-0.19.7.json", "Rust bindings for GStreamer"),
        ("human-panic-1.1.0.json", "Panic messages for humans"),
        ("hyperfine-1.15.0.json", "Command-line benchmarking tool"),
        ("iri-string-0.7.0.json", "IRI as string types"),
        ("libc-0.2.139.json", "Raw FFI bindings to platform libraries like libc"),
        ("predicates-2.1.5.json", "Implementation of boolean-valued predicate functions"),
        (
            "proc-macro2-1.0.50.json",
            "Substitute implementation of the compiler's proc_macro API to decouple token-based libraries from the procedural macro use case",
        ),
        ("quote-1.0.23.json", "Quasi-quoting macro quote!(...)"),
        ("rand-0.8.5.json", "Random number generators and other randomness functionality"),
        ("rand_core-0.6.4.json", "Core random number generator traits and tools for implementation"),
        ("regex-1.8.4.json", "Implementation of regular expressions for Rust"),
        ("regex-syntax-0.7.2.json", "Regular expression parser"),
        ("rpm-sequoia-1.2.0.json", "Implementation of the RPM PGP interface using Sequoia"),
        (
            "rust_decimal-1.28.0.json",
            "Decimal number implementation written in pure Rust suitable for financial and fixed-precision calculations",
        ),
        ("rustix-0.36.8.json", "Safe Rust bindings to POSIX/Unix/Linux/Winsock2-like syscalls"),
        ("serde-1.0.152.json", "Generic serialization/deserialization framework"),
        ("serde_derive-1.0.152.json", "Macros 1.1 implementation of #[derive(Serialize, Deserialize)]"),
        ("sha1collisiondetection-0.3.1.json", "SHA-1 hash function with collision detection and mitigation"),
        ("syn-1.0.107.json", "Parser for Rust source code"),
        ("time-0.3.17.json", "Date and time library"),
        ("tokio-1.25.0.json", "Event-driven, non-blocking I/O platform for writing asynchronous I/O backed applications"),
        (
            "unicode-xid-0.2.4.json",
            "Determine whether characters have the XID_Start or XID_Continue properties according to Unicode Standard Annex #31",
        ),
        ("zbus-3.8.0.json", "API for D-Bus communication"),
        ("zoxide-0.9.0.json", "Smarter cd command for your terminal"),
    ],
    ids=short_repr,
)
def test_metadata_get_summary(filename: str, expected: str):
    metadata = load_metadata_from_resource(filename)
    assert metadata.packages[0].get_summary() == expected
